# This program runs over a few different cutoff distances
# the proper cutoff may be estimated according to the sum of assignment scores
#
# Currently only the proton chemical shifts are used; involving the side chain
# 13C chemical shifts will be possible and may improve the performance.
#
# Here I only consider ILV labelling, but Ala and Met may help.
#
from sys import *
from string import *
import math

#------------------------------------------------------------------------------
# read in chemical shifts (predicted)
# they are categorized into I, L, and V
#
def read_shift(file,c_err,h_err):

  shifts = {}
  shifts['ALA'] = {}
  shifts['ILE'] = {}
  shifts['LEU'] = {}
  shifts['VAL'] = {}

  f = open(file,'r')
  lines = f.readlines()
  f.close()
  for line in lines:
    entry = line.split()
    if entry[1] not in shifts.keys():
      continue
    id = int(entry[0])
    res = entry[1]
    if id not in shifts[res].keys():
      shifts[res][id] = {}
    if entry[2][0] == 'C':
      shifts[res][id][entry[2]] = float(entry[3]) + c_err
    elif entry[2][0] == 'H':
      shifts[res][id][entry[2]] = float(entry[3]) + h_err

  return shifts

#------------------------------------------------------------------------------
# read the pdb file
#
def read_coor(file):

  atoms = {}
  f = open(file,'r')
  lines = f.readlines()
  f.close()
  m = 0
  for line in lines:
    entry = line.split()
    if entry[0] == 'TER':      # in case of polymer
      m = m + 1
      continue
    if entry[0] != 'ATOM':
      continue
    if not_methyl(entry[3],entry[2]):
      continue
    if entry[2][0] in '123':
      #id = (int(entry[4])+500*m, entry[3], entry[2][1:])
      id = (int(entry[4]), entry[3], entry[2][1:])
    else:
      #id = (int(entry[5])+500*m, entry[3], entry[2][:-1])
      id = (int(entry[5]), entry[3], entry[2][:-1])
    if id not in atoms.keys():
      atoms[id] = []
    if 'A' in entry[4]:
      coor = (float(entry[6]),float(entry[7]),float(entry[8]))
    else:
      coor = (float(entry[5]),float(entry[6]),float(entry[7]))
    atoms[id].append(coor)
  return atoms

#------------------------------------------------------------------------------
#
def not_methyl(res,atom):

  flag = 1
  if res == 'VAL' and 'HG' in atom:
    flag = 0
  if res == 'ALA' and 'HB' in atom:
    flag = 0
  if res == 'ILE' and 'HD' in atom:
    flag = 0
  if res == 'LEU' and 'HD' in atom:
    flag = 0
  return flag

#------------------------------------------------------------------------------
# compare the predicted and calculated chemical shifts
#
def compare(sh1, sh2, res):

  rms_ca = 0.98
  rms_cb = 1.10
  rms_hm = 0.13
  
  include = {}
  include['ILE'] = ('HD1',)
  include['LEU'] = ('HD1','HD2')
  include['VAL'] = ('HG1','HG2')

#------------------------------------------------------------------------------
# calculate distance
def calc_dist(atom1, atom2):

  dist = []
  for x1,y1,z1 in atom1:
    for x2,y2,z2 in atom2:
      d = math.sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2))
      dist.append(d)
  dist.sort()
  return dist[0]

#------------------------------------------------------------------------------
# simulate the expected NOE peaks
#
def noe_simulate(atoms,shift,assign,dist_l):

  category = ('ILE','LEU','VAL','ALA')
  include = {}
  include['ALA'] = ('HB',)
  include['ILE'] = ('HD1',)
  include['LEU'] = ('HD1','HD2')
  include['VAL'] = ('HG1','HG2')
  noes = {}
  #noes['ass'] = {}
  for aa in category:
    noes[aa] = {}
    cutoff = dist_l
    # peak from v/l is only half in theory
    if aa in ('VAL','LEU'):
      cutoff = 0.9*cutoff
    for id in shift[aa].keys():
      for atom in include[aa]:
        nid = (id,aa,atom)
        noes[aa][nid] = {}
        noes[aa][nid]['ass'] = {}
        sh1 = shift[aa][id][atom]
        coor = atoms[nid]
        for aa2 in category:
          noes[aa][nid][aa2] = []
          for id2 in shift[aa2].keys():
            if id2 == id:
              continue
            for atom2 in include[aa2]:
              nid2 = (id2,aa2,atom2)
              sh2 = shift[aa2][id2][atom2]
              coor2 = atoms[nid2]
              dist = calc_dist(coor,coor2)
              if dist > cutoff:
                continue
              # peak from v/l is only half in theory
              if aa2 in ('VAL','LEU') and dist > 0.9*cutoff:
                continue
              item = (nid2,sh2,dist)
              if nid2 in assign:
                noes[aa][nid]['ass'][nid2] = (sh2,dist)
              else:
                noes[aa][nid][aa2].append(item)
  return noes

#------------------------------------------------------------------------------
# read in experimental NOE from a formatted file
#
def read_noe(file,shift,assign):

  category = ('ILE','LEU','VAL','ALA')
  include = {}
  include['ALA'] = ('HB',)
  include['ILE'] = ('HD1',)
  include['LEU'] = ('HD1','HD2')
  include['VAL'] = ('HG1','HG2')
  noes = {}
  for aa in category:
    noes[aa] = {}
    for id in shift[aa].keys():
      for atom in include[aa]:
        if atom not in shift[aa][id].keys():
          continue
        nid = (id,aa,atom)
        noes[aa][nid] = {}
        noes[aa][nid]['ass'] = {}
        for aa2 in category:
          noes[aa][nid][aa2] = []
  f = open(file,'r')
  lines = f.readlines()
  f.close()
  for line in lines:
    entry = line.split()
    aa = entry[1]
    if aa not in category:
      continue
    id = int(entry[0])
    atom = entry[2]
    nid = (id,aa,atom)
    aa2 = entry[5]
    if aa2 not in category:
      continue
    id2 = int(entry[4])
    atom2 = entry[6]
    nid2 = (id2,aa2,atom2)
    item = (nid2, float(entry[7]), float(entry[-1]))
    if nid2 in assign.keys():
      nid3 = assign[nid2]
      noes[aa][nid]['ass'][nid3] = (item[1],item[2])
    else:
      noes[aa][nid][aa2].append(item)
  return noes

#------------------------------------------------------------------------------
# compare expected and observed NOE
#
def noe_match(noe1,noe2):

  sc_constant = 1.0
  rms_h = 0.13
  aa = ['ILE','LEU','VAL','ALA']
  score = 0
  for key in aa:
    list1 = noe1[key]
    list2 = noe2[key]
    if len(list1) == 0 or len(list2) == 0:
      continue
    l1 = len(list1)
    l2 = len(list2)
    if l1 > l2:
      count = l2
    else:
      count = l1
    score = score + sc_constant*count
    diff = abs(l1-l2)
    if count > 4:
      base = count
    else:
      base = 4
    if diff > 2:
      diff = float(diff)/base
      score = score - diff*diff
    sc1 = 0
    for p1 in list1:
      sh1 = p1[-2]
      diff = 5
      for p2 in list2:
        sh2 = p2[-2]
        if diff > abs((sh1-sh2)/rms_h):
          diff = abs((sh1-sh2)/rms_h)
      sc1 = sc1 + diff/5
    sc2 = 0
    for p2 in list2:
      sh2 = p2[-2]
      diff = 5
      for p1 in list1:
        sh1 = p1[-2]
        if diff > abs((sh1-sh2)/rms_h):
          diff = abs((sh1-sh2)/rms_h)
      sc2 = sc2 + diff/5
    if sc1 > sc2:
      score = score -sc1
    else:
      score = score - sc2
  for nid in noe1['ass'].keys():
    if nid in noe2['ass'].keys():
      score = score + 2.0
    else:
      score = score - 0.5

  return score

#------------------------------------------------------------------------------
#
def sorted_insert(list,item):

  l = len(list)
  if l == 0:
    list.append(item)
  else:
    for j in range(l):
      if item[-1] > list[j][-1]:
        list.insert(j,item)
        break
    if len(list) == l:
      list.append(item)

#------------------------------------------------------------------------------
# read in the available assignments
#
def read_assign(file):

  f = open(file,'r')
  lines = f.readlines()
  f.close()
  assigned = {}
  for line in lines:
    entry = line.split()
    if len(entry) == 0 or entry[0] == '#' or entry[0] == '?':
      continue
    nid1 = (int(entry[0]),entry[1],entry[2])
    nid2 = (int(entry[4]),entry[5],entry[6])
    assigned[nid1] = nid2
  return assigned

#------------------------------------------------------------------------------
# main

assigned = read_assign(argv[5])
p_assigned = assigned.keys()
r_assigned = []
for nid in p_assigned:
  nid2 = assigned[nid]
  r_assigned.append(nid2)
  print ' %6d %5s %5s -> %6d %5s %5s' % (nid[0],nid[1],nid[2],nid2[0],nid2[1],nid2[2])
print '========================================================'

shift1 = read_shift(argv[1],0,0)   # exp
shift2 = read_shift(argv[2],0,0)   # predict
atoms = read_coor(argv[3])           # structure
noe_exp = read_noe(argv[4],shift1,assigned)

category = ('ILE','LEU','VAL','ALA')
include = {}
include['ALA'] = ('HB',)
include['ILE'] = ('HD1',)
include['LEU'] = ('HD1','HD2')
include['VAL'] = ('HG1','HG2')

rms_hm = 0.13

for nn in range(21):
  cutoff = 6.0 + 0.1*nn
  outfile = 'mbp_ailv_%.1f' % cutoff
  f = open(outfile,'w')
  noe_pre = noe_simulate(atoms,shift2,r_assigned,cutoff)

  sum =0
  num_prot = 1
  num_correct = 0
  num_wrong = 0
  rel_correct = 0
  rel_wrong = 0
  for aa in category:
    all_list = []
    key1 = []
    for id1 in shift1[aa].keys():
      for atom1 in include[aa]:
        if atom1 not in shift1[aa][id1].keys():
          continue
        nid1 = (id1,aa,atom1)
        if nid1 in p_assigned:
          continue
        key1.append(nid1)
    key2 = []
    for id2 in shift2[aa].keys():
      for atom2 in include[aa]:
        if atom2 not in shift2[aa][id2].keys():
          continue
        nid2 = (id2,aa,atom2)
        if nid2 in r_assigned:
          continue
        key2.append(nid2)
    for nid1 in key1:
        id1 = nid1[0]
        atom1 = nid1[2]
        sh1 = shift1[aa][id1][atom1]
        str= '%6d %5s %5s:  %8.2f\n' % (id1, aa, atom1, sh1)
        f.write(str)
        noe1 = noe_exp[aa][nid1]
        rate_list = []
        for nid2 in key2:
            id2 = nid2[0]
            atom2 = nid2[2]
            sh2 = shift2[aa][id2][atom2]
            diff = abs(sh1-sh2)/rms_hm
            noe2 = noe_pre[aa][nid2]
            score = noe_match(noe1,noe2)-diff/5
            item = (nid2, score)
            sorted_insert(rate_list,item)
            item = (nid1, nid2, score)
            sorted_insert(all_list, item)
        for item in rate_list:
          id2, aa2, atom2 = item[0]
          str= '      %6d %5s %5s %8.2f\n' % (id2,aa2,atom2,item[1])
          f.write(str)
    f.write('\n')
    p_list = []
    r_list = []
    while len(p_list) < len(key1) and len(r_list) < len(key2):
      for nid1,nid2,sc in all_list:
        if nid1 in p_list or nid2 in r_list:
          continue
        p_list.append(nid1)
        r_list.append(nid2)
        sum = sum + sc
        for nid3,nid4,sc2 in all_list:
          if nid1 == nid3 and nid2 != nid4 or nid1 !=nid3 and nid2 == nid4:
            break
        if sc - sc2 < 0.2:
          if nid1[0] == nid2[0]:
            num_correct = num_correct + 1
            str= '? %6d %5s %5s -> %6d %5s %5s %8.2f\n' % (nid1[0],nid1[1],nid1[2],nid2[0],nid2[1],nid2[2],sc)
            f.write(str)
          else:
            num_wrong = num_wrong + 1
            str= '?x%6d %5s %5s -> %6d %5s %5s %8.2f\n' % (nid1[0],nid1[1],nid1[2],nid2[0],nid2[1],nid2[2],sc)
            f.write(str)
        else:
          if nid1[0] == nid2[0]:
            num_correct = num_correct + 1
            rel_correct = rel_correct + 1
            str= '  %6d %5s %5s -> %6d %5s %5s %8.2f\n' % (nid1[0],nid1[1],nid1[2],nid2[0],nid2[1],nid2[2],sc)
            f.write(str)
          else:
            num_wrong = num_wrong + 1
            rel_wrong = rel_wrong + 1
            str= 'x %6d %5s %5s -> %6d %5s %5s %8.2f\n' % (nid1[0],nid1[1],nid1[2],nid2[0],nid2[1],nid2[2],sc)
            f.write(str)
        break
    f.write( '========================================================\n')
  str = '%d, %d, %d, %d, %d, %.5f, %.1f\n' % (num_prot, num_correct, num_wrong, rel_correct, rel_wrong, sum, cutoff)
  f.write(str)
  f.close()
