# After the initial assignment of I, L, V, which has a correctness of >50%
# I think it's possible to look at the NOE connections: check whether the 
# expected NOE is observed according to the assignment
# If a residue is correctly assigned, I will expect that >50% NOE are observed
# Instead, if a residue is wrong, there may be many unexpected NOE and missing
# NOE
#
from sys import *
from string import *
import math

#------------------------------------------------------------------------------
# read in chemical shifts (predicted)
# they are categorized into I, L, and V
#
def read_shift(file,c_err,h_err):

  shifts = {}
  shifts['ALA'] = {}
  shifts['ILE'] = {}
  shifts['LEU'] = {}
  shifts['VAL'] = {}

  f = open(file,'r')
  lines = f.readlines()
  f.close()
  for line in lines:
    entry = line.split()
    if entry[1] not in shifts.keys():
      continue
    id = int(entry[0])
    res = entry[1]
    if id not in shifts[res].keys():
      shifts[res][id] = {}
    if entry[2][0] == 'C':
      shifts[res][id][entry[2]] = float(entry[3]) + c_err
    elif entry[2][0] == 'H':
      shifts[res][id][entry[2]] = float(entry[3]) + h_err

  return shifts

#------------------------------------------------------------------------------
# read the pdb file
#
def read_coor(file):

  atoms = {}
  f = open(file,'r')
  lines = f.readlines()
  f.close()
  m = 0
  for line in lines:
    entry = line.split()
    if entry[0] == 'TER':      # in case of polymer
      m = m + 1
      continue
    if entry[0] != 'ATOM':
      continue
    if not_methyl(entry[3],entry[2]):
      continue
    if entry[2][0] in '123':
      #id = (int(entry[4])+500*m, entry[3], entry[2][1:])
      id = (int(entry[4]), entry[3], entry[2][1:])
    else:
      #id = (int(entry[5])+500*m, entry[3], entry[2][:-1])
      id = (int(entry[5]), entry[3], entry[2][:-1])
    if id not in atoms.keys():
      atoms[id] = []
    if 'A' in entry[4]:
      coor = (float(entry[6]),float(entry[7]),float(entry[8]))
    else:
      coor = (float(entry[5]),float(entry[6]),float(entry[7]))
    atoms[id].append(coor)
  return atoms

#------------------------------------------------------------------------------
#
def not_methyl(res,atom):

  flag = 1
  if res == 'VAL' and 'HG' in atom:
    flag = 0
  if res == 'ALA' and 'HB' in atom:
    flag = 0
  if res == 'ILE' and 'HD' in atom:
    flag = 0
  if res == 'LEU' and 'HD' in atom:
    flag = 0
  return flag

#------------------------------------------------------------------------------
# calculate distance
def calc_dist(atom1, atom2):

  dist = []
  for x1,y1,z1 in atom1:
    for x2,y2,z2 in atom2:
      d = math.sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2))
      dist.append(d)
  dist.sort()
  return dist[0]

#------------------------------------------------------------------------------
# simulate the expected NOE peaks
#
def noe_simulate(atoms,shift,assign,dist_l):

  category = ('ILE','LEU','VAL','ALA')
  include = {}
  include['ALA'] = ('HB',)
  include['ILE'] = ('HD1',)
  include['LEU'] = ('HD1','HD2')
  include['VAL'] = ('HG1','HG2')
  noes = {}
  #noes['ass'] = {}
  for aa in category:
    noes[aa] = {}
    cutoff = dist_l
    # peak from v/l is only half in theory
    if aa in ('VAL','LEU'):
      cutoff = 0.9*cutoff
    for id in shift[aa].keys():
      for atom in include[aa]:
        nid = (id,aa,atom)
        noes[aa][nid] = {}
        noes[aa][nid]['ass'] = {}
        sh1 = shift[aa][id][atom]
        coor = atoms[nid]
        for aa2 in category:
          noes[aa][nid][aa2] = []
          for id2 in shift[aa2].keys():
            if id2 == id:
              continue
            for atom2 in include[aa2]:
              nid2 = (id2,aa2,atom2)
              sh2 = shift[aa2][id2][atom2]
              coor2 = atoms[nid2]
              dist = calc_dist(coor,coor2)
              if dist > cutoff:
                continue
              # peak from v/l is only half in theory
              if aa2 in ('VAL','LEU') and dist > 0.9*cutoff:
                continue
              item = (nid2,sh2,dist)
              if nid2 in assign:
                noes[aa][nid]['ass'][nid2] = (sh2,dist)
              else:
                noes[aa][nid][aa2].append(item)
  return noes

#------------------------------------------------------------------------------
# read in experimental NOE from a formatted file
#
def read_noe(file,shift,assign):

  category = ('ILE','LEU','VAL','ALA')
  include = {}
  include['ALA'] = ('HB',)
  include['ILE'] = ('HD1',)
  include['LEU'] = ('HD1','HD2')
  include['VAL'] = ('HG1','HG2')
  noes = {}
  for aa in category:
    noes[aa] = {}
    for id in shift[aa].keys():
      for atom in include[aa]:
        if atom not in shift[aa][id].keys():
          continue
        nid = (id,aa,atom)
        noes[aa][nid] = {}
        noes[aa][nid]['ass'] = {}
        for aa2 in category:
          noes[aa][nid][aa2] = []
  f = open(file,'r')
  lines = f.readlines()
  f.close()
  for line in lines:
    entry = line.split()
    aa = entry[1]
    if aa not in category:
      continue
    id = int(entry[0])
    atom = entry[2]
    nid = (id,aa,atom)
    aa2 = entry[5]
    if aa2 not in category:
      continue
    id2 = int(entry[4])
    atom2 = entry[6]
    nid2 = (id2,aa2,atom2)
    item = (nid2, float(entry[7]), float(entry[-1]))
    if nid2 in assign.keys():
      nid3 = assign[nid2]
      noes[aa][nid]['ass'][nid3] = (item[1],item[2])
    else:
      noes[aa][nid][aa2].append(item)
    if nid in assign.keys():
      nid3 = assign[nid]
      noes[aa2][nid2]['ass'][nid3] = (item[1],item[2])
  return noes

#------------------------------------------------------------------------------
#
def sorted_insert(list,item):

  l = len(list)
  if l == 0:
    list.append(item)
  else:
    for j in range(l):
      if item[-1] > list[j][-1]:
        list.insert(j,item)
        break
    if len(list) == l:
      list.append(item)

#------------------------------------------------------------------------------
# read in the available assignments
#
def read_assign(file):

  f = open(file,'r')
  lines = f.readlines()
  f.close()
  assigned = {}
  for line in lines:
    entry = line.split()
    if '->' not in entry:
      continue
    #if len(entry) == 0 or entry[0] == '#' or entry[0] == '?':
    #  continue
    if entry[0] in '?x' :
      nid1 = (int(entry[1]),entry[2],entry[3])
      nid2 = (int(entry[5]),entry[6],entry[7])
    else:
      nid1 = (int(entry[0]),entry[1],entry[2])
      nid2 = (int(entry[4]),entry[5],entry[6])
    assigned[nid1] = nid2
    #assigned[nid1] = nid1
  return assigned

#------------------------------------------------------------------------------
# main

assigned = read_assign(argv[5])
p_assigned = assigned.keys()
r_assigned = []
for nid in p_assigned:
  nid2 = assigned[nid]
  r_assigned.append(nid2)

shift1 = read_shift(argv[1],0,0)   # exp
shift2 = read_shift(argv[2],0,0.41)   # predict
atoms = read_coor(argv[3])           # structure
noe_exp = read_noe(argv[4],shift1,assigned)
noe_pre = noe_simulate(atoms,shift2,r_assigned,float(argv[6]))
rms_h = 0.13

network_list = []
for nid in p_assigned:
  nid2 = assigned[nid]
  noe1 = noe_exp[nid[1]][nid]['ass'].keys()
  noe2 = noe_pre[nid2[1]][nid2]['ass'].keys()
  print ' %6d %5s %5s -> %6d %5s %5s:' % (nid[0],nid[1],nid[2],nid2[0],nid2[1],nid2[2])
  match = 0
  extra = 0
  miss = 0
  for id1 in noe1:
    if id1 in noe2:
      match = match + 1
      print '              %6d %5s %5s' % id1
    else:
      extra = extra + 1
      print '    %6d %5s %5s' % id1
  for id2 in noe2:
    if id2 not in noe1:
      miss = miss + 1
      print '                               %6d %5s %5s' % id2
  print '%8d               %8d              %8d' % (extra,match,miss)
  if extra == 0 and miss == 0:
    reli = 2*match
  else:
    reli = float(match)/(extra+0.5*miss)
  sh1 = shift1[nid[1]][nid[0]][nid[2]]
  sh2 = shift2[nid2[1]][nid2[0]][nid2[2]]
  diff = abs(sh1-sh2)/rms_h
  item = (nid, miss, extra, match,reli-0.2*diff)
  sorted_insert(network_list,item)

print '#######'
for item in network_list:
  nid = item[0]
  nid2 = assigned[nid]
  if nid[0] == nid2[0]:
    print ' %6d %5s %5s -> %6d %5s %5s : %4d %4d %4d %4.1f' % (nid[0],nid[1],nid[2],nid2[0],nid2[1],nid2[2],item[1],item[2],item[3],item[4])
  else:
    print 'x%6d %5s %5s -> %6d %5s %5s : %4d %4d %4d %4.1f' % (nid[0],nid[1],nid[2],nid2[0],nid2[1],nid2[2],item[1],item[2],item[3],item[4])
